#pragma once
#include "vector.h"
#include "versor.h"


class Quaternion {
public:
    Vector im;
    Scalar re;

    Quaternion(const Vector& _im, Scalar _re): im(_im), re(_re) {}
    Quaternion(Scalar a, Scalar b, Scalar c, Scalar d): im(a,b,c), re(d) {}
    Quaternion(Scalar a): im(), re(a) {}

    void conjugate() { im = -im; }

    Quaternion operator-() const { return Quaternion(-im, -re); }
    Quaternion operator*(Scalar f) const { return Quaternion(im*f, re*f); }
    Quaternion operator/(Scalar f) const { return Quaternion(im/f, re/f); }
    Quaternion operator+(const Quaternion& o) const { return Quaternion(im+o.im, re+o.re); }


    Quaternion conjugated() const { return Quaternion(-im, re); }
    Quaternion operator~() const { return conjugated(); }

    static Quaternion axisAngle(const Versor& axis, Scalar angleRad) {
        return Quaternion(axis*sin(angleRad/2), cos(angleRad/2));
    }

    static Quaternion identity() {
        return Quaternion(0, 0, 0, 1);
    }

    // extract axis and angle from a quaternion encoding a roration
    Versor axis() const {
        return normalize(im);
    }

    // TODO:
    // Scalar angle() const

};

inline Quaternion operator*(const Quaternion& a, const Quaternion& b) {
    return Quaternion(
        a.re * b.im + a.im * b.re + cross(a.im, b.im),
        a.re * b.re - dot(a.im, b.im)
    );
}

/* this does the right thing and is quick to write... but can be optimized*/
inline Vector rotateUnoptimized(const Quaternion& q, const Vector& v) {
    return (q * Quaternion(v, 0) * ~q).im;
}

/* the optimized version of the above. Use this version! */
inline Vector rotate(const Quaternion& q, const Vector& v) {
    return dot(q.im, v) * q.im
           + (q.re * q.re) * v
           + (2 * q.re) * cross(q.im, v)
           - cross(cross(q.im, v), q.im);
}

inline Scalar length(const Quaternion& q) {
    return sqrt(dot(q.im, q.im) + q.re*q.re);
}

inline Quaternion normalize(const Quaternion& q) {
    return q/length(q);
}

inline Quaternion mix(const Quaternion& a, const Quaternion& b, Scalar t) {
    Scalar d = sign( dot(a.im, b.im) + a.re*b.re ); // shortest path!
    Quaternion res = a*(1-t) + b*t*d; // linear interpolation...
    return normalize(res);  // ...then re-normalized (NLERP)
}

inline bool isEqual(Quaternion q1, Quaternion q2) {
    return isEqual(q1.re, q2.re) && isEqual(q1.im, q2.im);
}

/* equivalent quaternions = they encode the same rotation */
inline bool isEquivalent(Quaternion q1, Quaternion q2) {
    return isEqual(q1, q2) || isEqual(q1, -q2);
}

